/*
 * This script prints multiple pages defined by a print configuration
 * and the therein referenced layer predefs (see ActivatePredef). A
 * print configuration is a simple text entity on the _DEFS layer (see
 * PredefTools) with the following lines:
 *         print
 *         MyName    <<config-name>>
 *         297       <<page-width in mm>>
 *         210       <<page-height in mm>>
 *                   >>empty line<<
 *         0.01      <<scale for page 1>>
 *         CELA1     <<name of layer to be used as centered bounding box for page 1>>
 *         PrNa1     <<name of predef for page 1>>
 *                   >>empty line<<
 *         0.1       <<scale for page 2>>
 *         CELA2     <<name of layer to be used as centered bounding box for page 2>>
 *         ...       >>repeat structure for every page to be printed<<
 * Optionally you may specify layer configurations starting with "@" (see ActivateLayerConfig)
 * resulting in something like
 *
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2018-11-09
 */

include("../MyTools.js");
include("../PredefTools.js");
include("scripts/File/Print/Print.js");

function MultiPrintToPdf(guiAction) {
    MyTools.call(this, guiAction);
}

MultiPrintToPdf.prototype = new MyTools();

MultiPrintToPdf.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);
    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();

    //get all configurations, the desired print configuration and the currently activated layers
    var predefs = new Array(), printConfigs = new Array(), layerConfigs = new Array();
    getAllPredefs(document, predefs, printConfigs, layerConfigs);
    if (predefs.length<1 || printConfigs.length<1) {
        EAction.handleUserMessage("predef layer not found or no text entities on it or no print configuration found");
        this.terminate();
        return;
    }
    var dialog = WidgetFactory.createWidget(
                   "scripts/MyTools",
                   "PredefSelectDialog.ui");
    var widgets = getWidgets(dialog);
    var list = widgets["list"];
    list.itemDoubleClicked.connect(dialog, "accept");
    for (var cd = 0; cd < printConfigs.length; cd++) {
        list.addItem(printConfigs[cd][0]);
    }
    list.setCurrentItem(list.item(0));
    if (!dialog.exec()) {
        this.terminate();
        return;
    }
    var oldLayers = getActiveLayers(document);

    //get global document configuration
    var config = printConfigs[list.currentRow];
    var configName = config[0];
    var width = config[1];
    var height = config[2];
    config.splice(0, 4);

    //get the target file name
    EAction.handleUserMessage("selected print configuration " + configName);
    var appWin = EAction.getMainWindow();
    var fileInfo = new QFileInfo(EAction.getDocument().getFileName());
    var filters = [ "PDF File (*.pdf)" ];
    var ret = File.getSaveFileName(appWin, qsTr("Multi-Export to PDF"),
                   fileInfo.absoluteFilePath(), filters);
    if (isNull(ret)) {
        this.terminate();
        return;
    }
    var pdfFile = ret[0];
    EAction.handleUserMessage("exporting to " + pdfFile);

    //prepare printing
    var mdiChild = EAction.getMdiChild();
    var view = mdiChild.getLastKnownViewWithFocus();
    var printer = new QPrinter(QPrinter.HighResolution);
    printer.setOutputFormat(QPrinter.PdfFormat);
    printer.setOutputFileName(pdfFile);
    printer.setPaperSize(new QSizeF(width, height));
    printer.setFullPage(true);
    printer.setOrientation(width > height ? QPrinter.Landscape : QPrinter.Portrait);
    printer.setColorMode(QPrinter.Color);
    var bgColor = view.getBackgroundColor();
    view.setBackgroundColor(new RColor("white"));
    view.setPrinting(true);
    var painter = new QPainter();
    painter.begin(printer);
    var unitScale = Print.getUnitScale(document);
    var widthInMM = printer.paperRect(QPrinter.Millimeter).width();
    var heightInMM = printer.paperRect(QPrinter.Millimeter).height();
    var paperBorder = new QRectF(0, 0, widthInMM, heightInMM);

    //for each page definition, print one page
    var page = 0, found;
    while (config.length >= 1) {
        if (config[0] === "") { //skip empty lines
            config.splice(0, 1);
        }
        else if (config[0].charAt(0) === "@") { //layer configuration
            var layerConfig = config[0].substr(1);
            config.splice(0, 1);

            //search desired layer config and activate it
            found = false;
            for (var ld = 0; ld < layerConfigs.length; ld++) {
                if (layerConfig === layerConfigs[ld][0]) {
                    found = true;
                    EAction.handleUserMessage("switching to layer configuration " + layerConfig);
                    restoreLayerConfiguration(document, di, layerConfigs[ld][1]);
                    break;
                }
            }
            if (!found) {
                EAction.handleUserMessage("could not find layer config " + predefName);
                break;
            }
        }
        else { //normal print configuration
            if (config.length < 3) { //incomplete configuration
                EAction.handleUserMessage("incomplete configuration");
                break;
            }

            //prepare page and current page config
            page++;
            if (page > 1) {
                printer.newPage();
            }
            var scale = config[0];
            var centerLayerName = config[1];
            var predefName = config[2];
            config.splice(0, 3);
            EAction.handleUserMessage("exporting page " + page + " with predef " + predefName + "...");

            //search desired predef and activate it
            found = false;
            for (var pd = 0; pd < predefs.length; pd++) {
                if (predefName === predefs[pd][0]) {
                    found = true;
                    activatePredefLayers(document, di, predefs[pd][1]);
                    break;
                }
            }
            if (!found) {
                EAction.handleUserMessage("could not find predef " + predefName);
                break;
            }
            activateAdditionalLayer(document, di, centerLayerName);

            //print the page
            printSinglePage(document, di, view, centerLayerName, printer, painter, paperBorder, unitScale, scale, width, height);
        }
    }

    //finish printing
    painter.end();
    printer.destroy();
    view.setBackgroundColor(bgColor);
    view.setPrinting(false);
    EAction.handleUserMessage("export done");
    activatePredefLayers(document, di, oldLayers);
    this.terminate();
};

/* this function was an idea to add the print configuration to the printed page, but it is not used at the moment
MultiPrintToPdf.prototype.drawText = function(document, painter, border, scale) {
    var font, fm, x, y, width, height;
    var path = new QPainterPath();
    var backgroundColor = new RColor("white");
    var str = "TestText";

    // border of visible area after gluing:
    var correctedBorder = border.translated(0,0);
    var top = correctedBorder.top();
    correctedBorder.setTop(correctedBorder.bottom());
    correctedBorder.setBottom(top);

    font = new QFont();
    font.fromString("Arial");
    font.setPointSizeF(10);

    var sf = 1.0;
    sf = RUnit.convert(sf, RS.Millimeter, document.getUnit()) / scale;
    path.addText(0, 0, font, str);

    // mirror text that is currently upside down:
    var t = new QTransform();
    t.scale(sf,-sf);
    path = t.map(path);

    width = path.boundingRect().width();
    height = path.boundingRect().height();

    x = correctedBorder.left() + 20/scale;
    y = correctedBorder.bottom();

    var color = new QColor("black");
    painter.setPen(new QPen(color));
    painter.setBrush(new QBrush(color));

    path.translate(x,y);
    painter.drawPath(path);
};*/

MultiPrintToPdf.init = function(basePath) {
    var action = new RGuiAction("&MultiPrintToPdf", RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/MultiPrintToPdf.js");
    action.setSortOrder(200);
    EAction.addGuiActionTo(action, MyTools, true, false, false);
};
