/*
 * This file contains helper functions for QCad scripts.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2012-04-06
 */

//this function fills the parameter arrays with the predefs and print configurations available in the given document on the _DEFS layer
//predefs is null or needs to be an already created array which is filled by this function (each entry is a two-element-array with name and layer list)
//printConfigs is null or needs to be an already created array which is filled by this function (each entry contains one print config list)
//layerConfigs is null or needs to be an already created array which is filled by this function (each entry contains one layer config list)
function getAllPredefs(document, predefs, printConfigs, layerConfigs) {
    var layer = document.queryLayer("_DEFS");
    if (isNull(layer)) { //the layer for the definitions is required
        return;
    }
    //get all entities of the definition layer and step through all text entities
    var entities = document.queryLayerEntities(layer.getId());
    for (var es = 0; es < entities.length; ++es) {
        var entity = document.queryEntityDirect(entities[es]);
        if (entity.getType() == RS.EntityText) {
            //all text entities on this layer are treated as
            //  - print configuration, if the first line matches "print" (see MultiPrintToPdf)
            //  - layer predef otherwise (see ActivatePredef)
            var lines = entity.getPlainText().split("\n");
            if (lines.length > 1 && lines[0].length > 0) {
                if (lines[0] === "print") { //print predef
                    if (!isNull(printConfigs)) { //add only if requested
                        lines.splice(0, 1);
                        printConfigs.push(lines);
                    }
                }
                else if (lines[0] === "layer") { //layer predef
                    if (!isNull(layerConfigs)) { //add only if requested
                        lines.splice(0, 1);
                        var curLayerConfig = new Array();
                        curLayerConfig[0] = lines[0];
                        lines.splice(0, 1);
                        curLayerConfig[1] = lines;
                        layerConfigs.push(curLayerConfig);
                    }
                }
                else { //normal predef
                    if (!isNull(predefs)) { //add only if requested
                        var curPredef = new Array();
                        curPredef[0] = lines[0];
                        lines.splice(0, 1);
                        curPredef[1] = lines;
                        predefs.push(curPredef);
                    }
                }
            }
        }
    }
    //sort the lists
    if (predefs !== null) predefs.sort(myCompare);
    if (printConfigs !== null) printConfigs.sort(myCompare);
    if (layerConfigs !== null) layerConfigs.sort(myCompare);
}

//helper function to sort the predef / config arrays
function myCompare(a, b) {
    if (a < b) {
        return -1;
    }
    if (a > b) {
       return  1;
    }
    return 0;
}

//returns a String array that contains all layer names of the given document
function getAllLayers(document) {
    var res = new Array();
    var layers = document.queryAllLayers();
    for (var l = 0; l < layers.length; ++l) {
        var layer = document.queryLayer(layers[l]);
        res.push(layer.getName());
    }
    res.sort();
    return res;
}

//return a String array that contains the names of all currently active (non-frozen) layers of the given document
function getActiveLayers(document) {
    var res = new Array();
    var layers = document.queryAllLayers();
    for (var l = 0; l < layers.length; ++l) {
        var layer = document.queryLayer(layers[l]);
        if (!layer.isFrozen()) {
            res.push(layer.getName());
        }
    }
    res.sort();
    return res;
}

//set all layers except those given in layerList to frozen, set all layers in layerList to active (non-frozen)
function activatePredefLayers(document, di, layerList) {
    var layers = document.queryAllLayers();
    var operation = new RModifyObjectsOperation();
    for (var l = 0; l < layers.length; ++l) {
        var layer = document.queryLayer(layers[l]);
        var frozen = true;
        for (var i = 0; i < layerList.length; ++i) {
            if (layerList[i] === layer.getName()) {
                frozen = false;
                break;
            }
        }
        layer.setFrozen(frozen);
        operation.addObject(layer);
    }
    di.applyOperation(operation);
    di.updateAllEntities();
    di.clearPreview();
    di.repaintViews();
}

//activate (set to non-frozen) one layer with a given name
function activateAdditionalLayer(document, di, layerName) {
    var layers = document.queryAllLayers();
    var operation = new RModifyObjectsOperation();
    for (var l = 0; l < layers.length; ++l) {
        var layer = document.queryLayer(layers[l]);
        if (layerName === layer.getName()) {
            layer.setFrozen(false);
            operation.addObject(layer);
            break;
        }
    }
    di.applyOperation(operation);
    di.updateAllEntities();
    di.clearPreview();
    di.repaintViews();
}

//set all properties of all given layers to specified configuration
function restoreLayerConfiguration(document, di, layerConfigList) {
    var lineweightList = RLineweight.getList(false);
    var operation = new RModifyObjectsOperation();
    for (var ci = 0; ci < layerConfigList.length; ci++) {
        var config = layerConfigList[ci].split("/");
        if (config.length === 4) {
            var layerName = config[0];
            var colorName = config[1];
            var lineweight = config[2];
            var linetype = config[3];
            var layer = document.queryLayer(document.getLayerId(layerName));
            layer.setColor(new RColor(colorName));
            layer.setLineweight(getLineweightByName(lineweightList, lineweight));
            layer.setLinetypeId(document.getLinetypeId(linetype));
            operation.addObject(layer);
        }
    }
    di.applyOperation(operation);
    di.updateAllEntities();
    di.clearPreview();
    di.repaintViews();
}

//helper function for restoreLayerConfiguration
function getLineweightByName(list, lname) {
    for (var i = 0; i < list.length; i++) {
        if (list[i][0] === lname) {
            return list[i][1];
        }
    }
    EAction.handleUserMessage("lineweight not found: " + lname);
    return null;
}

//print the currently active (non-frozen) layers to a single page sized width/height with prepared view, printer, painter, paperBorder, unitScale
//centerLayerName is used as bounding box and centered on the page, scale is completely independant
function printSinglePage(document, di, view, centerLayerName, printer, painter, paperBorder, unitScale, scale, width, height) {
    var layer = document.queryLayer(centerLayerName);
    var offset = null;
    if (isNull(layer)) {
        EAction.handleUserMessage("could not find center layer " + centerLayerName);
        break;
    }
    var entities = document.queryLayerEntities(layer.getId());
    var bBox = new RBox();
    for (var es = 0; es < entities.length; ++es) {
        var entity = document.queryEntityDirect(entities[es]);
        bBox.growToInclude(entity.getBoundingBox());
    }
    if (bBox.isValid()) {
        offset = bBox.getCenter().move(new RVector(-width/scale/2, -height/scale/2));
    }
    else {
        EAction.handleUserMessage("invalid center layer " + centerLayerName);
        break;
    }

    var destScale = unitScale * scale;
    var paperBorderTransformed = paperBorder.translated(0,0);
    paperBorderTransformed.setLeft(paperBorderTransformed.left()/destScale);
    paperBorderTransformed.setRight(paperBorderTransformed.right()/destScale);
    paperBorderTransformed.setTop(paperBorderTransformed.top()/destScale);
    paperBorderTransformed.setBottom(paperBorderTransformed.bottom()/destScale);
    paperBorderTransformed = paperBorderTransformed.translated(offset.x, offset.y);
    var printerFactor = new RVector(
        printer.paperRect().width() / width,
        printer.paperRect().height() / height
    );

    var transform = new QTransform();
    transform.scale(1, -1);
    transform.scale(printerFactor.x, printerFactor.y);
    transform.translate(-paperBorder.x(), -paperBorder.y()-paperBorder.height());
    transform.translate(-offset.x*destScale, -offset.y*destScale);
    transform.scale(destScale, destScale);
    painter.setWorldTransform(transform);

    view.setPrintPointSize(new RVector(1.0/destScale, 1.0/destScale));
    di.updateAllEntities();
    di.clearPreview();
    di.repaintViews();
    view.paintEntities(
        painter,
        new RBox(
            new RVector(paperBorderTransformed.x(), paperBorderTransformed.y()),
            new RVector(paperBorderTransformed.x()+paperBorderTransformed.width(), paperBorderTransformed.y()+paperBorderTransformed.height())
        )
    );

    var mi = bBox.getMinimum(), ma = bBox.getMaximum();
    painter.setPen(new QColor("white"));
    painter.setBrush(new QBrush(new QColor("white")));
    painter.drawRect(paperBorderTransformed.x(), paperBorderTransformed.y(), paperBorderTransformed.width(), mi.getY()-paperBorderTransformed.y());
    painter.drawRect(paperBorderTransformed.x(), ma.getY(), paperBorderTransformed.width(), mi.getY()-paperBorderTransformed.y());
    painter.drawRect(paperBorderTransformed.x(), paperBorderTransformed.y(), mi.getX()-paperBorderTransformed.x(), paperBorderTransformed.height());
    painter.drawRect(ma.getX(), paperBorderTransformed.y(), mi.getX()-paperBorderTransformed.x(), paperBorderTransformed.height());
    painter.setPen(new QPen(new QBrush(new QColor("black")), 0.2/scale, Qt.SolidLine, Qt.RoundCap, Qt.MiterJoin));
    painter.setBrush(Qt.NoBrush);
    painter.drawRect(mi.getX(), mi.getY(), ma.getX()-mi.getX(), ma.getY()-mi.getY());
}

