/*
 * This script moves + rotates + scales the current selection so that
 * the line built by the first two user picked points match
 * the line built by the last two user picked points. This may be
 * helpful if one needs to reconcile two figures.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2018-11-09
 */

include("../MyTools.js");
include("scripts/Modify/Transform.js");

function Reconcile(guiAction) {
    Transform.call(this, guiAction);

    this.adopt1 = undefined;
    this.adopt2 = undefined;
    this.target1 = undefined;
    this.target2 = undefined;
}

Reconcile.prototype = new Transform();

Reconcile.includeBasePath = includeBasePath;

Reconcile.State = {
    SettingAdoptPoint1 : 0,
    SettingAdoptPoint2 : 1,
    SettingTargetPoint1 : 2,
    SettingTargetPoint2 : 3
};

Reconcile.prototype.beginEvent = function() {
    Transform.prototype.beginEvent.call(this);

    if (!this.verifySelection()) {
        return;
    }

    this.copies = 0; //delete original
    this.useCurrentAttributes = false; //don't change to current attributes
    this.setState(Reconcile.State.SettingAdoptPoint1);
};

Reconcile.prototype.setState = function(state) {
    Transform.prototype.setState.call(this, state);

    this.getDocumentInterface().setClickMode(RAction.PickCoordinate);
    this.setCrosshairCursor();

    var appWin = RMainWindowQt.getMainWindow();
    var info;
    switch (this.state) {
    case Reconcile.State.SettingAdoptPoint1:
        this.adopt1 = undefined;
        this.adopt2 = undefined;
        this.target1 = undefined;
        this.target2 = undefined;
        info = qsTr("First point to adopt");
        appWin.setCommandPrompt(info);
        appWin.setLeftMouseTip(info);
        appWin.setRightMouseTip(EAction.trCancel);
        break;

    case Reconcile.State.SettingAdoptPoint2:
        this.adopt2 = undefined;
        this.target1 = undefined;
        this.target2 = undefined;
        info = qsTr("Second point to adopt");
        appWin.setCommandPrompt(info);
        appWin.setLeftMouseTip(info);
        appWin.setRightMouseTip(EAction.trBack);
        break;

    case Reconcile.State.SettingTargetPoint1:
        this.target1 = undefined;
        this.target2 = undefined;
        info = qsTr("First target point");
        appWin.setCommandPrompt(info);
        appWin.setLeftMouseTip(info);
        appWin.setRightMouseTip(EAction.trBack);
        break;

    case Reconcile.State.SettingTargetPoint2:
        this.target2 = undefined;
        info = qsTr("Second target point");
        appWin.setCommandPrompt(info);
        appWin.setLeftMouseTip(info);
        appWin.setRightMouseTip(EAction.trBack);
        break;
    }

    EAction.showSnapTools();
};

Reconcile.prototype.escapeEvent = function() {
    switch (this.state) {
    case Reconcile.State.SettingAdoptPoint1:
        Transform.prototype.escapeEvent.call(this);
        break;

    case Reconcile.State.SettingAdoptPoint2:
        this.setState(Reconcile.State.SettingAdoptPoint1);
        break;

    case Reconcile.State.SettingTargetPoint1:
        this.setState(Reconcile.State.SettingAdoptPoint2);
        break;

    case Reconcile.State.SettingTargetPoint2:
        this.setState(Reconcile.State.SettingTargetPoint1);
        break;
    }
};

Reconcile.prototype.pickCoordinate = function(event, preview) {
    var di = this.getDocumentInterface();
    var op;

    switch (this.state) {
    case Reconcile.State.SettingAdoptPoint1:
        if (!preview) {
            this.adopt1 = event.getModelPosition();
            di.setRelativeZero(this.adopt1);
            this.setState(Reconcile.State.SettingAdoptPoint2);
        }
        break;

    case Reconcile.State.SettingAdoptPoint2:
        this.adopt2 = event.getModelPosition();
        if (preview) {
            this.updatePreview();
        }
        else {
            this.setState(Reconcile.State.SettingTargetPoint1);
        }
        break;

    case Reconcile.State.SettingTargetPoint1:
        this.target1 = event.getModelPosition();
        if (preview) {
            this.updatePreview();
        }
        else {
            di.setRelativeZero(this.target1);
            this.setState(Reconcile.State.SettingTargetPoint2);
        }
        break;

    case Reconcile.State.SettingTargetPoint2:
        this.target2 = event.getModelPosition();
        if (preview) {
            this.updatePreview();
        }
        else {
            op = this.getOperation(false);
            di.applyOperation(op);
            di.setRelativeZero(this.target1);
            this.setState(-1);
            this.terminate();
        }
        break;
    }
};

Reconcile.prototype.getOperation = function(preview) {
    if (!isValidVector(this.adopt1) || !isValidVector(this.adopt2)
            || !isValidVector(this.target1) || !isValidVector(this.target2)) {
        return undefined;
    }

    //see http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_vector.html
    var moveX, moveY, adoptCon, targetCon;
    adoptCon = new RLine(this.adopt1, this.adopt2);
    targetCon = new RLine(this.target1, this.target2);
    moveX = this.target1.getX() - this.adopt1.getX();
    moveY = this.target1.getY() - this.adopt1.getY();
    this.moveBy = new RVector(moveX, moveY);
    this.angle = targetCon.getAngle() - adoptCon.getAngle();
    this.scale = targetCon.getLength() / adoptCon.getLength();
    if (!preview) {
        EAction.handleUserMessage("move by " + moveX + " / " + moveY + ", rotate by " + (this.angle * 180 / Math.PI) + "°, scale by " + this.scale);
    }
    return Transform.prototype.getOperation.call(this, preview);
};

Reconcile.prototype.transform = function(entity, k, op) {
    entity.move(this.moveBy);
    entity.rotate(this.angle, this.target1);
    entity.scale(this.scale, this.target1);
    op.addObject(entity, this.useCurrentAttributes);
};

Reconcile.prototype.getAuxPreview = function() {
    var ret = undefined;
    if (isValidVector(this.adopt1) && isValidVector(this.adopt2)) {
        ret = new Array();
        ret.push(new RLine(this.adopt1, this.adopt2));
        if (isValidVector(this.target1) && isValidVector(this.target2)) {
            ret.push(new RLine(this.target1, this.target2));
        }
    }
    return ret;
};

Reconcile.init = function(basePath) {
    var action = new RGuiAction(qsTr("&Reconcile"), 
        RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setRequiresSelection(true);
    action.setScriptFile(basePath + "/Reconcile.js");
    action.setIcon(basePath + "/Reconcile.svg");
    action.setSortOrder(500);
    EAction.addGuiActionTo(action, MyTools, true, false, false);
    var appWin = EAction.getMainWindow();
    appWin.addSelectionListener(action);
};
