/*
 * This script acts as a widget which reduces the current selection
 * depending on an user entered entity property.
 *
 * Copyright smf, some ideas taken from the QCad example scripts,
 *   from Widgtes/PropertyEditor there was much copy+paste
 * last change: 2019-05-02
 */

include("scripts/WidgetFactory.js");

/**
 * Internal helper class. Notified when properties are changed.
 */
function SWPropertyWatcher(selectWidget, sender, propertyType) {
    this.selectWidget = selectWidget;
    this.sender = sender;
    this.propertyType = propertyType;
}

/**
 * Called when a property has been changed by the user. Triggers
 * an updateSelection.
 */
SWPropertyWatcher.prototype.propertyChanged = function(value) {
    var propTitle = this.propertyType.getPropertyTitle();
    //check general property
    switch (propTitle) {
        case "Layer":
            this.selectWidget.updateSelection(function(e) {
                return e.getLayerName();
            }, value, false);
            return;
        case "Color":
            this.selectWidget.updateSelection(function(e) {
                return e.getColor().getName();
            }, this.sender.itemData(value).getName(), false);
            return;
        case "Lineweight":
            this.selectWidget.updateSelection(function(e) {
                return RLineweight.getName(e.getLineweight());
            }, RLineweight.getName(this.sender.itemData(value)), false);
            return;
        case "Linetype":
            this.selectWidget.updateSelection(function(e) {
                return e.getLinetypeId();
            }, EAction.getDocument().getLinetypeId(this.sender.itemData(value).getName()), false);
            return;
    }

    //all other properties
    var near = false;
    var propId = this.propertyType.getId();
    if (this.sender.toString().startsWith("QComboBox")) {
        value = this.sender.itemData(value);
    }
    else if (this.sender.toString() === "QLineEdit") { //value is string from a line edit (e.g. text contents)
        value = this.sender.text;
        if (value === this.propertyEditor.varies) {
            return;
        }
    }
    else if (this.sender.toString().startsWith("RMathLineEdit")) { //value is number from a math line edit
        if (this.sender.text === this.sender.originalText) {
            return;
        }
        this.sender.setProperty("originalText", this.sender.text);
        value = this.sender.getValue();
        if (isNaN(value)) {
            this.propertyEditor.updateGui(true);
            return;
        }
        near = true;
    }
    var pt = new RPropertyTypeId(propId);
    this.selectWidget.updateSelection(function(e) {
        return e.getProperty(pt)[0];
    }, value, near);
    return;
};




/**
 * Internal helper class. Notified when indices of list properties are changed.
 */
function IndexWatcher(selectWidget, sender, propertyType) {
    this.selectWidget = selectWidget;
    this.sender = sender;
    this.propertyType = propertyType;
}

/**
 * Called when the index of a list property has changed and the list value
 * needs to be changed to show the list entry at the new index.
 */
IndexWatcher.prototype.indexChanged = function(index) {
    var objectName = SelectWidgetImpl.getControlObjectName(this.propertyType);
    var control = this.selectWidget.widget.findChild(objectName);
    this.selectWidget.initNumberControls(objectName, this.propertyType, true, control, index);

    //debugger;

    var indexControlObjectName =
        SelectWidgetImpl.getIndexControlObjectName(this.propertyType.getPropertyGroupTitle());
    var indexControl = this.selectWidget.widget.findChild(indexControlObjectName);

    if (isNull(indexControl.controlNames)) {
        return;
    }

    var di = EAction.getDocumentInterface();
    if (isNull(di)) {
        return;
    }

    var controlX = this.selectWidget.widget.findChild(indexControl.controlNames[0]);
    var controlY = this.selectWidget.widget.findChild(indexControl.controlNames[1]);

    if (isNull(controlX) || isNull(controlY)) {
        return;
    }

    var x = controlX.getValue();
    var y = controlY.getValue();

    var view = di.getLastKnownViewWithFocus();
    if (isNull(view)) {
        return;
    }

    var r = view.mapDistanceFromView(5);

    di.clearPreview();
    di.addAuxShapeToPreview(new RCircle(new RVector(x, y), r));
    di.repaintViews();
};



/**
 * \class SelectWidgetImpl
 * \brief Property editor implementation.
 * \ingroup ecma_widgets
 */
function SelectWidgetImpl(basePath) {
    RPropertyEditor.call(this);

    this.entityTypeFilter = RS.EntityAll;
    this.varies = qsTr("*VARIES*");
    this.widget = WidgetFactory.createWidget(basePath, "SelectWidget.ui");
    this.basePath = basePath;

    this.colon = qsTr(":");

    this.widget.findChild("LabelLayer").text = RSettings.translate("REntity", "Layer") + this.colon;
    this.widget.findChild("LabelColor").text = RSettings.translate("REntity", "Color") + this.colon;
    this.widget.findChild("LabelLineweight").text = RSettings.translate("REntity", "Lineweight") + this.colon;
    this.widget.findChild("LabelLinetype").text = RSettings.translate("REntity", "Linetype") + this.colon;

    var typeCombo = this.widget.findChild("Type");
    typeCombo["activated(int)"].connect(this, "typeChanged");
    typeCombo.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
    typeCombo.focusPolicy = Qt.ClickFocus;

    // initialize fixed general properties at the top:
    var layerCombo = this.widget.findChild("Layer");
    layerCombo['activated(QString)'].connect(
                new SWPropertyWatcher(this, layerCombo, REntity.PropertyLayer),
                'propertyChanged');
    layerCombo.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
    layerCombo.focusPolicy = Qt.ClickFocus;

    var colorCombo = this.widget.findChild("Color");
    colorCombo['activated(int)'].connect(
                new SWPropertyWatcher(this, colorCombo, REntity.PropertyColor),
                'propertyChanged');
    colorCombo.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
    colorCombo.focusPolicy = Qt.ClickFocus;

    var lineweightCombo = this.widget.findChild("Lineweight");
    lineweightCombo['activated(int)'].connect(
                new SWPropertyWatcher(this, lineweightCombo, REntity.PropertyLineweight),
                'propertyChanged');
    lineweightCombo.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
    lineweightCombo.focusPolicy = Qt.ClickFocus;

    var linetypeCombo = this.widget.findChild("Linetype");
    linetypeCombo['activated(int)'].connect(
                new SWPropertyWatcher(this, linetypeCombo, REntity.PropertyLinetype),
                'propertyChanged');
    linetypeCombo.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
    linetypeCombo.focusPolicy = Qt.ClickFocus;

    this.geometryGroup = undefined;
    this.customGroup = undefined;
    this.onlyChangesOverride = undefined;
}

SelectWidgetImpl.prototype = new RPropertyEditor();

/**
 * Here the selection is updated
 */
SelectWidgetImpl.prototype.updateSelection = function(leftFunc, comp, near) {
//    EAction.handleUserMessage("updateSelection for " + deref + " compare to " + comp);
    var doc = EAction.getDocument();
    if (isNull(doc)) {
        return;
    }
    var di = EAction.getDocumentInterface();
    var reduceToNotBy = this.widget.findChild("ReduceTo")['checked'];
    if (!this.widget.findChild("ReduceNear")['checked']) near = false; //disable if not requested
    var entityIds = doc.querySelectedEntities();
    var deList = [];
    for (var es = 0; es < entityIds.length; ++es) {
        var id = entityIds[es];
//        EAction.handleUserMessage("calling...");
        var left = leftFunc(doc.queryEntity(id));
//        EAction.handleUserMessage("returned");
        if (reduceToNotBy ?
                ( near ? Math.abs(left-comp) > Math.abs(comp*0.001) : left !== comp ) :
                ( near ? Match.abs(left-comp) < Math.abs(comp*0.001) : left === comp )) {
//            EAction.handleUserMessage("deselecting " + id + " with " + left + " compared to " + comp);
            deList.push(id);
        }
    }
    di.deselectEntities(deList);
};

/**
 * Implementation from RPropertyEditor to update the property editor GUI.
 */
SelectWidgetImpl.prototype.updateGui = function(onlyChanges, entityTypeFilter) {
    var row;
//    if (!isNull(entityTypeFilter)) {
//        this.entityTypeFilter=entityTypeFilter;
//    }

//    if (isNull(this.entityTypeFilter)) {
//        debugger;
//    }

    if (!isNull(this.onlyChangesOverride)) {
        onlyChanges = this.onlyChangesOverride;
        this.onlyChangesOverride = undefined;
    }

    this.widget.updatesEnabled = false;
    if (!onlyChanges) {
        if (!isNull(this.geometryGroup)) {
            this.geometryGroup.destroy();
            this.geometryGroup = undefined;
        }
        if (!isNull(this.childGroup)) {
            this.childGroup.destroy();
            this.childGroup = undefined;
        }
        if (!isNull(this.customGroup)) {
            this.customGroup.destroy();
            this.customGroup = undefined;
        }
    }

    var typeCombo = this.widget.findChild("Type");

    var generalGroup = this.widget.findChild("General");
    var layerCombo = this.widget.findChild("Layer");
    var colorCombo = this.widget.findChild("Color");
    var lineweightCombo = this.widget.findChild("Lineweight");
    var linetypeCombo = this.widget.findChild("Linetype");

    var scrollArea = this.widget.findChild("ScrollArea");
    var layout = scrollArea.layout();

    if (!onlyChanges) {
        typeCombo.clear();
    }

    var groups = this.getGroupTitles();

    // no properties to show:
    if (groups.length===0) {
        typeCombo.insertItem(0, qsTr("No Selection"));
        typeCombo.setEnabled(false);
        layerCombo.clear();
        colorCombo.currentIndex = 0;
        lineweightCombo.currentIndex = 0;
        linetypeCombo.currentIndex = 0;
        generalGroup.enabled = false;
        this.widget.updatesEnabled = true;
        return;
    }

    var gridLayoutGeometry = undefined;
    var gridLayoutCustom = undefined;

    if (!onlyChanges) {
        // create geometry group box with grid layout:
        this.geometryGroup = new QGroupBox(qsTr("Geometry"), this.widget);
        layout.insertWidget(3, this.geometryGroup);

        // grid layout with three columns and N rows for N property controls:
        gridLayoutGeometry = new QGridLayout(this.geometryGroup);
        gridLayoutGeometry.setVerticalSpacing(4);
        gridLayoutGeometry.setColumnStretch(0,0);
        gridLayoutGeometry.setColumnStretch(1,1);
        gridLayoutGeometry.setColumnStretch(2,0);
        this.geometryGroup.setLayout(gridLayoutGeometry);

        // custom properties:
        if (RSettings.isXDataEnabled()) {
            // create custom property group box with grid layout:
            this.customGroup = new QGroupBox(qsTr("Custom"), this.widget);
            layout.insertWidget(3, this.customGroup);

            // grid layout with four columns and N rows for N property controls:
            // (the fourth column is for the 'remove property' button)
            gridLayoutCustom = new QGridLayout(this.customGroup);
            gridLayoutCustom.setVerticalSpacing(4);
            gridLayoutCustom.setColumnStretch(0,0);
            gridLayoutCustom.setColumnStretch(1,1);
            gridLayoutCustom.setColumnStretch(2,0);
            gridLayoutCustom.setColumnStretch(3,0);
            this.customGroup.setLayout(gridLayoutCustom);
        }
    }

    var firstEntry = true;

    // for all property groups:
    for (var gi=0; gi<groups.length; ++gi) {
        var group = groups[gi];
        //qDebug("group: ", group);

        // for all properties in that group:
        var titles = this.getPropertyTitles(group);
        for (var pi=0; pi<titles.length; ++pi) {
            var title = titles[pi];
            //qDebug("title: ", title);
            var value = this.getPropertyValue(group, title);
            var attributes = this.getPropertyAttributes(group, title);
            //var propertyTypeId = attributes.getPropertyTypeId();
            var propertyTypeId = RPropertyTypeId.getPropertyTypeId(group, title);
            if (isNull(value)) {
                // invalid value means no property found:
                continue;
            }
//            qDebug("\tproperty: ", title, ": ", value);
//            qDebug("\tproperty type id: ", propertyTypeId);

//            if (attributes.hasEnumChoices()) {
//                var ec = attributes.getEnumChoices();
//                qDebug("enumChoices: ", ec.length);
//                qDebug("enumChoices: ", ec[0]);
//                qDebug("enumChoices: ", ec[1]);
//                qDebug("enumChoices: ", ec[2]);
//                debugger;
//            }

            // ignore entity type:
            if (propertyTypeId.getId()===REntity.PropertyType.getId()) {
                // ignore
            }

            // ignore block ID:
            else if (propertyTypeId.getId()===REntity.PropertyBlock.getId()) {
                // ignore
            }

            // layer:
            else if (propertyTypeId.getId()===REntity.PropertyLayer.getId()) {
                this.initControls(propertyTypeId, onlyChanges, layerCombo);
            }

            // color:
            else if (propertyTypeId.getId()===REntity.PropertyColor.getId()) {
                this.initControls(propertyTypeId, onlyChanges, colorCombo);
            }

            // lineweight:
            else if (propertyTypeId.getId()===REntity.PropertyLineweight.getId()) {
                this.initControls(propertyTypeId, onlyChanges, lineweightCombo);
            }

            // linetype:
            else if (propertyTypeId.getId()===REntity.PropertyLinetype.getId()) {
                this.initControls(propertyTypeId, onlyChanges, linetypeCombo);
            }

            // geometric properties:
            else {
                var gridLayout;
                var groupBox = undefined;
                if (propertyTypeId.isCustom()) {
                    // block reference attributes:
                    if (propertyTypeId.getCustomPropertyTitle()==="Attributes") {
                        gridLayout = gridLayoutChild;
                        groupBox = this.childGroup;
                        // only show group with dependent entities if needed:
                        if (!isNull(this.childGroup)) {
                            this.childGroup.visible = true;
                        }
                    }
                    // custom properties:
                    else {
                        gridLayout = gridLayoutCustom;
                        groupBox = this.customGroup;
                    }
                }
                // geometric properties:
                else {
                    gridLayout = gridLayoutGeometry;
                    groupBox = this.geometryGroup;
                }

                // for the first property of a new property group,
                // add group title and index controls if the property is a list:
                if (pi===0) {
                    //debugger;
                    // create group label:
                    if (!onlyChanges && group.length!==0 && !attributes.isList()) {
                        var groupLabel = new QLabel(RSettings.translate("REntity", group), groupBox);
                        groupLabel.styleSheet = "margin-bottom:0px;font-weight:bold;";
                        if (!firstEntry) {
                            groupLabel.styleSheet += "margin-top:0px;";
                            firstEntry = false;
                        }
                        gridLayout.addWidget(groupLabel, gridLayout.rowCount(),0, 1,3);
                    }
                }

                // don't display handle and draw order
                var controls;
                if (title==="Handle" || title==="Draw Order") {
                    controls = undefined;
                }
                else {
                    controls = this.initControls(propertyTypeId, onlyChanges);
                }

                // add property name and controls to layout:
                if (!onlyChanges) {
                    if (!isNull(controls)) {
                        row = gridLayout.rowCount();
                        var localTitle = title;
                        if (propertyTypeId.isCustom() && group!==RSettings.getAppId() && propertyTypeId.getCustomPropertyTitle()!=="Attributes") {
                            var pos = title.indexOf("_");
                            if (pos!==-1) {
                                localTitle = title.mid(pos+1);
                            }
                        }
                        else {
                            localTitle = RSettings.translate("REntity", title);
                        }

                        var l = attributes.getLabel();
                        if (l.length>0) {
                            localTitle = l;
                        }

                        var label = new QLabel(localTitle + this.colon, groupBox);
                        label.textInteractionFlags = Qt.TextSelectableByMouse;
                        label.alignment = Qt.AlignRight | Qt.AlignVCenter;

                        gridLayout.addWidget(label, row,0);
                        if (controls.length===1) {
                            gridLayout.addWidget(controls[0], row,1, 1,2);
                        }
                        else if (controls.length===2) {
                            gridLayout.addWidget(controls[0], row,1);
                            gridLayout.addWidget(controls[1], row,2);
                        }
                        else {
                            debugger;
                        }

                        firstEntry = false;
                    }
                }
            }
        }
    }

    // update type combo box at the top for entity filters:
    if (!onlyChanges) {
        var types = this.getTypes();
        var totalCount = 0;
        for (var ti=0; ti<types.length; ti++) {
            var type = types[ti];
            var typeCount = this.getTypeCount(type);
            totalCount += typeCount;

            //qDebug("type: ", type, " / count: ", typeCount);
            typeCombo.addItem(entityTypeToString(type) + " (" + typeCount + ")", type);
        }
        //if (totalCount!=1) {
        if (types.length!=1) {
            typeCombo.insertItem(0, qsTr("All") + " (" + totalCount + ")", RS.EntityAll);
            typeCombo.setEnabled(true);
        }
        else {
            typeCombo.setEnabled(false);
        }
        var index = typeCombo.findData(this.getEntityTypeFilter());
        if (index===-1) {
            typeCombo.currentIndex = 0;
        }
        else {
            typeCombo.currentIndex = index;
        }

        generalGroup.enabled = true;
        this.geometryGroup.enabled = true;
    }

    this.widget.updatesEnabled = true;
};

/**
 * Initializes a control for the given property.
 *
 * \param propertyTypeId The propery the returned control edits.
 * \param onlyChanges Control already exists and only need to be re-initialized.
 * \param control The control to initialize if it is already known by the caller.
 *
 * \return The control(s) as array.
 */
SelectWidgetImpl.prototype.initControls = function(propertyTypeId, onlyChanges, control) {
    var value = this.getAdjustedPropertyValue(propertyTypeId);
    var attributes = this.getPropertyAttributes(propertyTypeId);
    var controls;

    var objectName = SelectWidgetImpl.getControlObjectName(propertyTypeId);
    if (isNull(control) && onlyChanges) {
        // find existing control:
        control = this.widget.findChild(objectName);
        if (isNull(control)) {
            // no control available for this property (e.g. fit point of spline with control points):
            return undefined;
        }
    }

    // Horizontal alignment: combo box:
    if (propertyTypeId.getId()===RTextEntity.PropertyHAlign.getId()) {
        controls = this.initChoiceControls(
                    objectName, propertyTypeId, onlyChanges, control,
                    [ qsTr("Left"), qsTr("Center"), qsTr("Right"),
                      qsTr("Aligned"), qsTr("Middle"), qsTr("Fit")],
                    [ RS.HAlignLeft, RS.HAlignCenter, RS.HAlignRight,
                      RS.HAlignAlign, RS.HAlignMid, RS.HAlignFit]);
    }

    // Vertical alignment: combo box:
    else if (propertyTypeId.getId()===RTextEntity.PropertyVAlign.getId()) {
        controls = this.initChoiceControls(
                    objectName, propertyTypeId, onlyChanges, control,
                    [qsTr("Top"), qsTr("Middle"), qsTr("Base"), qsTr("Bottom")],
                    [RS.VAlignTop, RS.VAlignMiddle, RS.VAlignBase, RS.VAlignBottom]);
    }

    // Hatch pattern: combo box with hatch names:
    else if (propertyTypeId.getId()===RHatchEntity.PropertyPatternName.getId()) {
        var patternNames = undefined;
        if (RUnit.isMetric(EAction.getDocument().getUnit())) {
            patternNames = RPatternListMetric.getNames();
        }
        else {
            patternNames = RPatternListImperial.getNames();
        }

        controls = this.initChoiceControls(
                    objectName, propertyTypeId, onlyChanges, control,
                    patternNames);
    }

    // Font: combo box with font names:
    else if (propertyTypeId.getId()===RTextEntity.PropertyFontName.getId()) {
        controls = this.initChoiceControls(
                    objectName, propertyTypeId, onlyChanges, control,
                    patternNames);
    }

    // List:
    else if (getTypeOf(value)==="array") {
        //not supported: list of properties
        return undefined;
    }

    // Number: math line edit:
    else if (!isComboBox(control) && typeof(value)==="number") {
        // TODO: handle int / double differently:
        controls = this.initNumberControls(objectName, propertyTypeId, onlyChanges, control);
    }

    // Boolean: combo box with Yes, No, *VARIES*:
    else if (typeof(value)==="boolean") {
        controls = this.initBooleanControls(objectName, propertyTypeId, onlyChanges, control);
    }

    // Choices: combo box:
    else if (attributes.hasChoices() || isComboBox(control)) {
        controls = this.initChoiceControls(objectName, propertyTypeId, onlyChanges, control);
    }

    // String: line edit:
    else if (typeof(value)==="string") {
        controls = this.initStringControls(objectName, propertyTypeId, onlyChanges, control);
        if (!onlyChanges) {
            if (attributes.isDimensionLabel()) {
                WidgetFactory.initLineEdit(controls[0], true);
                controls[0].placeholderText = qsTr("Auto");
            }
            else if (attributes.isLabel()) {
                WidgetFactory.initLineEdit(controls[0], false);
            }
        }
    }

    if (isNull(controls) || controls.length===0) {
        return undefined;
    }

    control = controls[0];

    if (attributes.isReadOnly()) {
        control.readOnly = true;
        this.makeGray(control);
    }

    return controls;
};

/**
 * Initializes a number control (RMathLineEdit) for the given property.
 * \internal
 */
SelectWidgetImpl.prototype.initNumberControls = function(objectName, propertyTypeId, onlyChanges, control, index) {
    var value = this.getAdjustedPropertyValue(propertyTypeId);
    if (isNumber(index)) {
        value = value[index];
    }
    var attributes = this.getPropertyAttributes(propertyTypeId);

    if (isNull(control)) {
        control = new RMathLineEdit(this.geometryGroup);
        if (attributes.isAngleType()) {
            control.setAngle(true);
        }
        control.objectName = objectName;
    }

    var newText;
    if (attributes.isMixed()) {
        control.setEnabled(true);
        newText = this.varies;
    }
    else {
        control.setEnabled(false);
        if (attributes.isAngleType()) {
            value = RMath.rad2deg(value);
        }
        var document = EAction.getDocument();
        if (document.getLinearFormat()===RS.Fractional || document.getLinearFormat()===RS.FractionalStacked) {
            newText = RUnit.getLabel(value, document, true);
        }
        else {
            newText = sprintf("%.6f", value);
        }
    }

    if (control.text !== newText) {
        control.text = newText;
        control.setProperty("originalText", newText);
        if (isFunction(control.clearError)) {
            control.clearError();
        }
    }

    if (attributes.isInteger()) {
        control.setInteger(true);
    }

    if (!onlyChanges) {
        control.editingFinished.connect(
                    new SWPropertyWatcher(this, control, propertyTypeId),
                    'propertyChanged');
    }

    return new Array(control);
};

/**
 * Initializes a string control (QLineEdit) for the given property.
 * \internal
 */
SelectWidgetImpl.prototype.initStringControls = function(objectName, propertyTypeId, onlyChanges, control) {
    var value = this.getAdjustedPropertyValue(propertyTypeId);
    var attributes = this.getPropertyAttributes(propertyTypeId);
    //var editAsRichTextButton;
    var clearButton = undefined;

    if (isNull(control)) {
        control = new QLineEdit(value, this.geometryGroup);
        control.objectName = objectName;
        /*if (attributes.isRichText()) {
            editAsRichTextButton = new QToolButton(this.geometryGroup);
            editAsRichTextButton.iconSize = new QSize(12,12);
            var action = RGuiAction.getByScriptFile("scripts/Modify/EditText/EditText.js");
            action.setProperty("");
            editAsRichTextButton.setDefaultAction(action);
        }*/
    }

    if (attributes.isMixed()) {
        control.setEnabled(true);
        control.text = this.varies;
        // TODO: make sure that the start of the text is shown:
        //control.home(false);
    }
    else {
        control.setEnabled(false);
        if (control.text!=value) {
            control.text = value;
        }
    }

    if (!onlyChanges) {
        control.editingFinished.connect(
                    new SWPropertyWatcher(this, control, propertyTypeId),
                    'propertyChanged');
    }

    if (isNull(clearButton)) {
        return new Array(control);
    }
    else {
        return new Array(control, clearButton);
    }
};

/**
 * Initializes a boolean control (combo box with yes / no) for the given property.
 * \internal
 */
SelectWidgetImpl.prototype.initBooleanControls = function(objectName, propertyTypeId, onlyChanges, control) {
    var value = this.getAdjustedPropertyValue(propertyTypeId);
    var attributes = this.getPropertyAttributes(propertyTypeId);

    if (isNull(control)) {
        control = new QComboBox(this.geometryGroup);
        control.objectName = objectName;
        control.addItem(qsTr("Yes"), true);
        control.addItem(qsTr("No"), false);
        control.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
        control.focusPolicy = Qt.ClickFocus;
    }

    if (attributes.isMixed()) {
        control.setEnabled(true);
        if (!onlyChanges) {
            control.insertItem(0, this.varies, this.varies);
        }
        control.currentIndex = 0;
    }
    else {
        control.setEnabled(false);
        // not varies anymore:
        var variesIndex = control.findText(this.varies);
        if (variesIndex!==-1) {
            control.removeItem(variesIndex);
        }
        control.currentIndex = control.findData(value);
    }

    if (!onlyChanges) {
        control['activated(int)'].connect(
                    new SWPropertyWatcher(this, control, propertyTypeId),
                    'propertyChanged');
    }

    return new Array(control);
};

/**
 * Initializes a choice control (combo box) for the given property.
 * \internal
 */
SelectWidgetImpl.prototype.initChoiceControls = function(
    objectName, propertyTypeId, onlyChanges, control, choices, choicesData) {
    
    var i;

    var value = this.getAdjustedPropertyValue(propertyTypeId);
    var attributes = this.getPropertyAttributes(propertyTypeId);

    if (isNull(control)) {
        if (propertyTypeId.getId()===RTextEntity.PropertyFontName.getId()) {
//            control = new QComboBox(this.geometryGroup);
            control = new QFontComboBox(this.geometryGroup);
            //control.sizeAdjustPolicy = QComboBox.AdjustToContentsOnFirstShow;
            control.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Preferred);
            //control.maxWidth = 100;
            control.editable = false;
            control.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
            control.focusPolicy = Qt.ClickFocus;
            initFontComboBox(control);
        }
        else {
            control = new QComboBox(this.geometryGroup);
            control.installEventFilter(new REventFilter(QEvent.Wheel.valueOf(), true));
            control.focusPolicy = Qt.ClickFocus;
        }

        control.objectName = objectName;

        if (isNull(choicesData)) {
            control['activated(QString)'].connect(
                        new SWPropertyWatcher(this, control, propertyTypeId),
                        'propertyChanged');
        }
        else {
            control['activated(int)'].connect(
                        new SWPropertyWatcher(this, control, propertyTypeId),
                        'propertyChanged');
        }
    }

    if (isOfType(control, QComboBox)) {
        control.clear();
        if (isNull(choices)) {
            control.addItems(attributes.getChoices());
            control.model().sort(0);
        }
        else {
            if (isNull(choicesData)) {
                control.addItems(choices);
            }
            else {
                for (i=0; i<choices.length && i<choicesData.length; ++i) {
                    control.addItem(choices[i], choicesData[i]);
                }
            }
        }
    }

    // mixed:
    var variesIndex = control.findText(this.varies);
    if (attributes.isMixed()) {
        control.setEnabled(true);
        if (variesIndex!=-1) {
            control.currentIndex = variesIndex;
        }
        else {
            control.insertItem(0, this.varies, this.varies);
            control.currentIndex = 0;
        }
        return new Array(control);
    }

    control.setEnabled(false);
    // not mixed anymore, remove mixed option:
    if (variesIndex!=-1) {
        control.removeItem(variesIndex);
    }

    var index = -1;
    if (typeof(value)==="string") {
        index = control.findText(value);
        if (index===-1) {
            control.addItem(value);
            index = control.findText(value);
        }
    }
    else {
        for (i=0; i<control.count; ++i) {
            var data = control.itemData(i);
            var match = false;

            // for some types (e.g. RColor), we have to use equals here:
            if (!isNull(data) && isFunction(data.equals)) {
                if (data.equals(value)) {
                    match = true;
                }
            }
            else {
                if (data===value) {
                    match = true;
                }
            }

            if (match) {
                index = i;
                break;
            }
        }
    }

    if (index===-1) {
        // handle custom color:
        if (isOfType(control, RColorCombo)) {
            control.setColor(value);
        }
        else {
            // custom something, cannot handle this:
            debugger;
        }
    }
    else {
        control.currentIndex = index;
    }

    return new Array(control);
};

SelectWidgetImpl.getControlObjectName = function(propertyTypeId) {
    if (propertyTypeId.isCustom()) {
        return "Control" + propertyTypeId.getCustomPropertyName();
    }
    else {
        return "Control" + propertyTypeId.getId();
    }
};

SelectWidgetImpl.getIndexControlObjectName = function(group) {
    return "IndexControl" + group.replace(" ", "_");
};

/**
 * Called when user changes entity filter at the top of the property editor.
 */
SelectWidgetImpl.prototype.typeChanged = function() {
    var typeCombo = this.widget.findChild("Type");
    this.updateSelection(function(e) {
        return e.getType();
    }, typeCombo.itemData(typeCombo.currentIndex), false);
};

SelectWidgetImpl.prototype.getAdjustedPropertyValue = function(propertyTypeId) {
    // show handle as hex reprensentation (string):
    if (propertyTypeId.getId()===RObject.PropertyHandle.getId()) {
        var value = this.getPropertyValue(propertyTypeId);
        if (value===RObject.INVALID_HANDLE) {
            return "-";
        }
        return sprintf("0x%x", value);
    }

    return this.getPropertyValue(propertyTypeId);
};

/**
 * Make read-only controls more distinguished.
 */
SelectWidgetImpl.prototype.makeGray = function(control) {
    if (isNull(control)) {
        return;
    }

    var p = control.palette;
    p.setColor(QPalette.Base, new QColor("#eeeeee"));
    control.palette = p;
};

function SelectWidget(guiAction) {
    Widgets.call(this, guiAction);
}

SelectWidget.prototype = new Widgets();

/**
 * Shows / hides the property editor.
 */
SelectWidget.prototype.beginEvent = function() {
    Widgets.prototype.beginEvent.call(this);

    var appWin = RMainWindowQt.getMainWindow();
    var dock = appWin.findChild("SelectWidgetDock");
    if (!QCoreApplication.arguments().contains("-no-show")) {
        dock.visible = !dock.visible;
    }
};

SelectWidget.prototype.finishEvent = function() {
    Widgets.prototype.finishEvent.call(this);

    var appWin = RMainWindowQt.getMainWindow();
    var dock = appWin.findChild("SelectWidgetDock");
    this.getGuiAction().setChecked(dock.visible);
};

SelectWidget.init = function(basePath) {
    var appWin = EAction.getMainWindow();

    var action = new RGuiAction(qsTr("Select Widget"), appWin);
    action.setRequiresDocument(false);
    action.setScriptFile(basePath + "/SelectWidget.js");
    action.setIcon(basePath + "/SelectWidget.svg");
    action.setSortOrder(11001);
    EAction.addGuiActionTo(action, MyTools, true, true, false);

    var pe = new SelectWidgetImpl(basePath);
    var dock = new RDockWidget(qsTr("Select Widget"), appWin);
    dock.objectName = "SelectWidgetDock";
    dock.setWidget(pe.widget);
    appWin.addPropertyListener(pe);
    appWin.addDockWidget(Qt.RightDockWidgetArea, dock);
    SelectWidget.instance = pe;

    dock.shown.connect(function() { action.setChecked(true); });
    dock.hidden.connect(function() { action.setChecked(false); });
};
